﻿from pyevolve import Util
from pyevolve import GTree
from pyevolve import GSimpleGA
from pyevolve import Consts
from pyevolve import DBAdapters
import pydot
import math
import os
import multiprocessing

rmse_accum = Util.ErrorAccumulator()

@GTree.gpdec(representation="+", color="red", shape="circle")
def gp_add(a, b): return a+b

@GTree.gpdec(representation="-", color="red", shape="circle")
def gp_sub(a, b): return a-b

@GTree.gpdec(representation="*", color="chartreuse3", shape="circle")
def gp_mul(a, b): return a*b

@GTree.gpdec(representation="/", color="chartreuse3", shape="circle")
def gp_div(a, b): return a/b if b!=0 else 0

@GTree.gpdec(representation="%", color="chartreuse3", shape="circle")
def gp_mod(a, b): return a%b if b!=0 else 1

@GTree.gpdec(representation="||x||", color="gold", shape="circle")
def gp_round(a): return round(a)

@GTree.gpdec(representation="|x|", color="gold", shape="circle")
def gp_abs(a): return abs(a)

@GTree.gpdec(representation="c", color="gold", shape="circle")
def gp_ceil(a): return math.ceil(a)

@GTree.gpdec(representation="f", color="gold", shape="circle")
def gp_floor(a): return math.floor(a)

@GTree.gpdec(representation="√", color="cornflowerblue", shape="circle")
def gp_sqrt(a):   return math.sqrt(abs(a))

@GTree.gpdec(representation="ln", color="cornflowerblue", shape="circle")
def gp_ln(a):   return math.log(a) if a>0 else 0

#@GTree.gpdec(representation="x^y", color="cornflowerblue", shape="circle")
#def gp_pow(a,b):   return math.pow(a,b)

primes=[]

def load_primes():
	with open('primes.txt') as dat:
		global primes
		primes = [int(x) for x in dat.read().split(' ')]
		#primes=primes[:100]

def eval_func(chromosome):
	global rmse_accum
	rmse_accum.reset()
	code_comp = chromosome.getCompiledCode()
	
	#pr = chromosome.getParam("primes")
	
	for n in range(1,100):
		evaluated     = eval(code_comp)
		target        = primes[n-1]
		rmse_accum   += (target, evaluated)
	
	return rmse_accum.getRMSE()
	
def evolve_callback(ga_engine):
	generation = ga_engine.getCurrentGeneration()
	if generation%50==0:
		indiv = ga_engine.bestIndividual()
		print ''.join(map(chr,indiv))
	return False

def main_run():
	genome = GTree.GTreeGP()
	
	#global primes
	#params = {'max_depth':4, 'method':"ramped", 'primes':primes}

	#enome.setParams(rangemin=min(numeric_sentence),
    #                rangemax=max(numeric_sentence),
    #                bestrawscore=0.00,
    #                gauss_mu=1, gauss_sigma=4)
	
	params = {'max_depth':4, 'method':"ramped"}
	#genome.setParams(max_depth=4, method="ramped")
	genome.setParams(**params)
	genome.evaluator += eval_func
	

	#genome.crossover.set(Crossovers.G1DListCrossoverCutCrossfill)

	#genome.mutator.set(Mutators.GTreeGPMutatorSubtree)
	#ga = GSimpleGA.GSimpleGA(genome, seed=666)
	ga = GSimpleGA.GSimpleGA(genome)

	ga.setParams(gp_terminals = ['n', 'ephemeral:random.randint(-10,10)', 'ephemeral:random.random()'], gp_function_prefix = "gp")

	ga.setMinimax(Consts.minimaxType["minimize"])
	ga.setGenerations(50)
	ga.setCrossoverRate(1.0)
	ga.setMutationRate(0.05)
	ga.setPopulationSize(500)
	ga.setElitism(True)
	#ga.setMultiProcessing(True)
	
	#ga.stepCallback.set(evolve_callback)
	#ga.terminationCriteria.set(GSimpleGA.RawScoreCriteria)
	
	#csv_adapter = DBAdapters.DBFileCSV(identify="run1", filename="stats.csv",frequency = 10, reset = True)
	#ga.setDBAdapter(csv_adapter)
	
	#vpython_adapter = DBAdapters.DBVPythonGraph(identify="primes", frequency=1)
	#ga.setDBAdapter(vpython_adapter)
	
	sqlite_adapter = DBAdapters.DBSQLite(identify="ex1")
	ga.setDBAdapter(sqlite_adapter)
	
	ga(freq_stats=10)
	best = ga.bestIndividual()
	print "\nExpression: " + best.getPreOrderExpression() + "\n"
	
	for n in range(1,21):
		num = eval(best.getPreOrderExpression())
		print str(primes[n-1]) + "\t" + str(num)

	print "\nSoulution fitness: " + str(best.getRawScore()) + "\n"
		
	graph = pydot.Dot()
	ga.bestIndividual().writeDotGraph(graph)
	graph.write_png('tree.png', prog='dot')

	os.system("start tree.png");

if __name__ == "__main__":
	load_primes()
	main_run()
