﻿from pyevolve import Util
from pyevolve import GTree
from pyevolve import GSimpleGA
from pyevolve import Consts
from pyevolve import DBAdapters
import pydot
import math
import os
import multiprocessing

import datetime
import numpy as np

import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt

import sqlite3
import subprocess

rmse_accum = Util.ErrorAccumulator()

@GTree.gpdec(representation="+", color="red", shape="circle")
def gp_add(a, b): return a+b

@GTree.gpdec(representation="-", color="red", shape="circle")
def gp_sub(a, b): return a-b

@GTree.gpdec(representation="*", color="chartreuse3", shape="circle")
def gp_mul(a, b): return a*b

@GTree.gpdec(representation="/", color="chartreuse3", shape="circle")
def gp_div(a, b): return a/b if b!=0 else 0

@GTree.gpdec(representation="%", color="chartreuse3", shape="circle")
def gp_mod(a, b): return a%b if b!=0 else 1

#@GTree.gpdec(representation="||x||", color="gold", shape="circle")
#def gp_round(a): return round(a)

#@GTree.gpdec(representation="|x|", color="gold", shape="circle")
#def gp_abs(a): return abs(a)

#@GTree.gpdec(representation="c", color="gold", shape="circle")
#def gp_ceil(a): return math.ceil(a)

#@GTree.gpdec(representation="f", color="gold", shape="circle")
#def gp_floor(a): return math.floor(a)

@GTree.gpdec(representation="√", color="cornflowerblue", shape="circle")
def gp_sqrt(a):   return math.sqrt(abs(a))

@GTree.gpdec(representation="ln", color="cornflowerblue", shape="circle")
def gp_ln(a):   return math.log(a) if a>0 else 0

#@GTree.gpdec(representation="x^y", color="cornflowerblue", shape="circle")
#def gp_pow(a,b):   return math.pow(a,b)

primes=[]

def load_primes():
	with open('primes.txt') as dat:
		global primes
		primes = [int(x) for x in dat.read().split(' ')]

def plot_difference(run, expression):
	graph = plt.subplot(111)

	x = np.arange(0, 20, 1)
	
	global primes
	y = primes[:20]	# real primes
	y2 = [int(round(eval(expression))) for n in range(1,21)] # predicted primes
	y3=[(j-i) for i,j in zip(y,y2)]	# diference

	graph.set_ylabel('p(n)')
	graph.set_xlabel('n')
	graph.set_title(u'stvarni / predviđeni prosti brojevi')

	p1, = graph.plot(x, y, 'r-', lw=2, alpha=0.7)
	p2, = graph.plot(x, y2, 'g-', lw=2, alpha=0.7)
	p3, = graph.plot(x, y3, 'b-', lw=2, alpha=0.7)

	graph.grid(True)
	#graph.grid(color='gray', ls='--', lw=0.5, which='both', alpha=0.5)
	graph.legend([p1, p2, p3], ['stvarno', u'predviđeno', 'razlika'], loc=2)

	#plt.show()
	plt.plot()
	plt.savefig('img/' + run + '_difference')
	plt.close()

def eval_func(chromosome):
	global rmse_accum
	global primes
	rmse_accum.reset()
	code_comp = chromosome.getCompiledCode()
	
	test_set = chromosome.getParam('test_set')
	test_set_step = chromosome.getParam('test_set_step')
	
	for n in range(1,test_set+1,test_set_step):
		evaluated     = int(round(eval(code_comp)))
		target        = primes[n-1]
		rmse_accum   += (target, evaluated)
	
	return rmse_accum.getRMSE()
	
def eval_func2(chromosome):
	global primes
	code_comp = chromosome.getCompiledCode()
	
	test_set = chromosome.getParam('test_set')
	test_set_step = chromosome.getParam('test_set_step')
	
	sum = 0
	for n in range(1,test_set+1,test_set_step):
		evaluated     = int(round(eval(code_comp)))
		target        = primes[n-1]
		sum   += 0 if target==evaluated else 100 if abs(target-evaluated)<5 else 10
	
	return sum

def genetic_run(iter, run_id,tree_depth,crossover_rate,population_size,test_set,test_set_step):
	genome = GTree.GTreeGP()

	params = {'max_depth':tree_depth, 'method':"ramped", 'bestrawscore':0, 'test_set':test_set, 'test_set_step':test_set_step}
	genome.setParams(**params)
	genome.evaluator += eval_func2

	#genome.crossover.set(Crossovers.G1DListCrossoverCutCrossfill)
	#genome.mutator.set(Mutators.GTreeGPMutatorSubtree)
	
	ga = GSimpleGA.GSimpleGA(genome)
	#ga = GSimpleGA.GSimpleGA(genome, seed=666)

	ga.setParams(gp_terminals = ['n', 'ephemeral:random.randint(-10,10)', 'ephemeral:random.random()'], gp_function_prefix = "gp")

	# constant params
	ga.setMinimax(Consts.minimaxType["minimize"])
	ga.setGenerations(50)
	#ga.setElitism(True)
	#ga.setMultiProcessing(True)
	
	# variable params
	ga.setCrossoverRate(crossover_rate)
	ga.setMutationRate(0.2)
	ga.setPopulationSize(population_size)

	# run string
	run = 'run' +  str(run_id) + "_" + str(iter+1)

	# sqlite adapter
	sqlite_adapter = DBAdapters.DBSQLite(identify=run)
	ga.setDBAdapter(sqlite_adapter)

	# run ga
	ga(freq_stats=10)
	best = ga.bestIndividual()
	print "\nExpression: " + best.getPreOrderExpression()
	print '\nSoulution fitness (RMSE fitness): ' + str(best.getFitnessScore()) + '\n'
	
	# save run to database
	print 'Saving to database...\n'
	conn = sqlite3.connect('autorun.db')
	c = conn.cursor()
	c.execute('insert into runs values (%s,%s,%s,"%s",%s,%s,%s,%s,%s)' % (run_id,iter,best.getFitnessScore(),best.getPreOrderExpression(),tree_depth,crossover_rate,population_size,test_set,test_set_step))
	conn.commit()
	global t0
	c.execute('update stats set current_run=%s, total_runs=%s, total_time="%s" where id=1' % (run_id,120,str(datetime.datetime.now() - t0)))
	conn.commit()
	c.close()
	
	# generate AST graph
	print 'Generating AST graph...\n'
	graph = pydot.Dot()
	ga.bestIndividual().writeDotGraph(graph)
	graph.write_png('img/' + run + '_ast.png', prog='dot')
	
	# generate fitness score graph
	print 'Generating fitness graph...\n'
	theproc = subprocess.Popen('pyevolve_graph.py -3 -i' + run + ' -m -o' + 'img/' + run + '_fitness.png', shell = True)
	theproc.communicate()  
	#subprocess.call(['python pyevolve_graph.py -4 -i run1_1 -m'])
	
	# generate true/predicted difference graph
	print 'Generating difference graph...\n'
	plot_difference(run, best.getCompiledCode())

t0=datetime.datetime.now()
	
def main_run():

	run = 0
	t0 = datetime.datetime.now()
	
	tree_depth = 6
	crossover_rate = 0.8;
	
	#for tree_depth in range(9,3,-1):
	#	for crossover_rate in np.arange(0.7, 1.01, 0.1):
	for population_size in [2000]:
		for test_set in [20]:
			for test_set_step in [1]:
				if test_set == 10 and test_set_step==10: continue
				run+=1
				print '\n-----------------------------------------------------------------------------------------\nRUN ' + str(run) + '\n-----------------------------------------------------------------------------------------'
				print 'Total run time: ' + str(datetime.datetime.now() - t0)
				print 'Runs remaining: ' + str(25-run+1) + '\n'
				
				for iter in range(0,25):
					print 'Starting run iteration %s with parameters: %s / %s / %s / %s / %s / %s\n' % (iter,run,tree_depth,crossover_rate,population_size,test_set,test_set_step)
					
					genetic_run(iter,run,tree_depth,crossover_rate,population_size,test_set,test_set_step)

if __name__ == "__main__":
	load_primes()
	main_run()
